"""
Command line interface for pygount.
"""

# Copyright (c) 2016-2024, Thomas Aglassinger.
# All rights reserved. Distributed under the BSD License.
import argparse
import contextlib
import logging
import os
import sys

from rich.progress import Progress

import pygount
import pygount.analysis
import pygount.common
import pygount.write

#: Valid formats for option --format.
VALID_OUTPUT_FORMATS = ("cloc-xml", "json", "sloccount", "summary")

_DEFAULT_ENCODING = "automatic"
_DEFAULT_OUTPUT_FORMAT = "sloccount"
_DEFAULT_OUTPUT = "STDOUT"
_DEFAULT_SOURCE_PATTERNS = os.curdir
_DEFAULT_SUFFIXES = "*"

_HELP_ENCODING = '''encoding to use when reading source code; use "automatic"
 to take BOMs, XML prolog and magic headers into account and fall back to
 UTF-8 or CP1252 if none fits; use "automatic;<fallback>" to specify a
 different fallback encoding than CP1252; use "chardet" to let the chardet
 package determine the encoding; default: "%(default)s"'''

_HELP_EPILOG = """SHELL-PATTERN is a pattern using *, ? and ranges like [a-z]
 as placeholders. PATTERNS is a comma separated list of SHELL-PATTERN. The
 prefix [regex] indicated that the PATTERNS use regular expression syntax. If
 default values are available, [...] indicates that the PATTERNS extend the
 existing default values."""

_HELP_FORMAT = 'output format, one of: {}; default: "%(default)s"'.format(
    ", ".join(['"' + output_format + '"' for output_format in VALID_OUTPUT_FORMATS])
)

_HELP_GENERATED = """comma separated list of regular expressions to detect
 generated code; default: %(default)s"""

_HELP_MERGE_EMBEDDED_LANGUAGES = """merge counts for embedded languages into
 their base language; for example, HTML+Jinja2 counts as HTML"""

_HELP_FOLDERS_TO_SKIP = """comma separated list of glob patterns for folder
 names not to analyze. Use "..." as first entry to append patterns to the
 default patterns; default: %(default)s"""

_HELP_NAMES_TO_SKIP = """comma separated list of glob patterns for file names
 not to analyze. Use "..." as first entry to append patterns to the default
 patterns; default: %(default)s"""

_HELP_SUFFIX = '''limit analysis on files matching any suffix in comma
 separated LIST; shell patterns are possible; example: "py,sql"; default:
 "%(default)s"'''

_OUTPUT_FORMAT_TO_WRITER_CLASS_MAP = {
    "cloc-xml": pygount.write.ClocXmlWriter,
    "json": pygount.write.JsonWriter,
    "sloccount": pygount.write.LineWriter,
    "summary": pygount.write.SummaryWriter,
}
assert set(VALID_OUTPUT_FORMATS) == set(_OUTPUT_FORMAT_TO_WRITER_CLASS_MAP.keys())

_log = logging.getLogger("pygount")


def _check_encoding(name, encoding_to_check, alternative_encoding, source=None):
    """
    Check that ``encoding`` is a valid Python encoding
    :param name: name under which the encoding is known to the user, e.g. 'default encoding'
    :param encoding_to_check: name of the encoding to check, e.g. 'utf-8'
    :param source: source where the encoding has been set, e.g. option name
    :raise pygount.common.OptionError if ``encoding`` is not a valid Python encoding
    """
    assert name is not None

    if encoding_to_check not in (alternative_encoding, "chardet", None):
        try:
            "".encode(encoding_to_check)
        except LookupError:
            raise pygount.common.OptionError(
                f'{name} is "{encoding_to_check}" but must be "{alternative_encoding}" or a known Python encoding',
                source,
            ) from None


class Command:
    """
    Command interface for pygount, where options starting with defaults can
    gradually be set and finally :py:meth:`execute()`.
    """

    def __init__(self):
        self.set_encodings(_DEFAULT_ENCODING)
        self._folders_to_skip = pygount.common.regexes_from(pygount.analysis.DEFAULT_FOLDER_PATTERNS_TO_SKIP_TEXT)
        self._generated_regexs = pygount.common.regexes_from(pygount.analysis.DEFAULT_GENERATED_PATTERNS_TEXT)
        self._has_duplicates = False
        self._has_summary = False
        self._has_to_merge_embedded_languages = False
        self._is_verbose = False
        self._names_to_skip = pygount.common.regexes_from(pygount.analysis.DEFAULT_NAME_PATTERNS_TO_SKIP_TEXT)
        self._output = _DEFAULT_OUTPUT
        self._output_format = _DEFAULT_OUTPUT_FORMAT
        self._source_patterns = _DEFAULT_SOURCE_PATTERNS
        self._suffixes = pygount.common.regexes_from(_DEFAULT_SUFFIXES)

    def set_encodings(self, encoding, source=None):
        encoding_is_chardet = (encoding == "chardet") or (encoding.startswith("chardet;"))
        if encoding_is_chardet and not pygount.analysis.has_chardet:  # pragma: no cover
            raise pygount.common.OptionError('chardet must be installed to set default encoding to "chardet"')
        if encoding in ("automatic", "chardet"):
            default_encoding = encoding
            fallback_encoding = None
        elif encoding.startswith(("automatic;", "chardet;")):
            first_encoding_semicolon_index = encoding.find(";")
            default_encoding = encoding[:first_encoding_semicolon_index]
            fallback_encoding = encoding[first_encoding_semicolon_index + 1 :]
        else:
            default_encoding = encoding
            fallback_encoding = pygount.analysis.DEFAULT_FALLBACK_ENCODING
        self.set_default_encoding(default_encoding, source)
        self.set_fallback_encoding(fallback_encoding, source)

    @property
    def default_encoding(self):
        return self._default_encoding

    def set_default_encoding(self, default_encoding, source=None):
        _check_encoding("default encoding", default_encoding, "automatic", source)
        self._default_encoding = default_encoding

    @property
    def fallback_encoding(self):
        return self._fallback_encoding

    def set_fallback_encoding(self, fallback_encoding, source=None):
        _check_encoding("fallback encoding", fallback_encoding, "automatic", source)
        self._fallback_encoding = fallback_encoding

    @property
    def folders_to_skip(self):
        return self._folders_to_skip

    def set_folders_to_skip(self, regexes_or_patterns_text, source=None):
        self._folders_to_skip = pygount.common.regexes_from(
            regexes_or_patterns_text, pygount.analysis.DEFAULT_FOLDER_PATTERNS_TO_SKIP_TEXT, source
        )

    @property
    def generated_regexps(self):
        return self._generated_regexs

    def set_generated_regexps(self, regexes_or_patterns_text, source=None):
        self._generated_regexs = pygount.common.regexes_from(
            regexes_or_patterns_text, pygount.analysis.DEFAULT_GENERATED_PATTERNS_TEXT, source
        )

    @property
    def has_duplicates(self):
        return self._has_duplicates

    def set_has_duplicates(self, has_duplicates, source=None):
        self._has_duplicates = bool(has_duplicates)

    @property
    def has_to_merge_embedded_languages(self):
        return self._has_to_merge_embedded_languages

    def set_has_to_merge_embedded_languages(self, has_to_merge_embedded_languages, source=None):
        self._has_to_merge_embedded_languages = bool(has_to_merge_embedded_languages)

    @property
    def is_verbose(self):
        return self._is_verbose

    def set_is_verbose(self, is_verbose, source=None):
        self._is_verbose = bool(is_verbose)

    @property
    def names_to_skip(self):
        return self._names_to_skip

    def set_names_to_skip(self, regexes_or_pattern_text, source=None):
        self._names_to_skip = pygount.common.regexes_from(
            regexes_or_pattern_text, pygount.analysis.DEFAULT_NAME_PATTERNS_TO_SKIP_TEXT, source
        )

    @property
    def output(self):
        return self._output

    def set_output(self, output, source=None):
        assert output is not None
        self._output = output

    @property
    def output_format(self):
        return self._output_format

    def set_output_format(self, output_format, source=None):
        assert output_format is not None
        if output_format not in VALID_OUTPUT_FORMATS:
            raise pygount.common.OptionError(
                f"format is {output_format} but must be one of: {VALID_OUTPUT_FORMATS}", source
            )
        self._output_format = output_format

    @property
    def source_patterns(self):
        return self._source_patterns

    def set_source_patterns(self, glob_patterns_or_text, source=None):
        assert glob_patterns_or_text is not None
        self._source_patterns = pygount.common.as_list(glob_patterns_or_text)
        assert len(self._source_patterns) >= 0

    @property
    def suffixes(self):
        return self._suffixes

    def set_suffixes(self, regexes_or_patterns_text, source=None):
        assert regexes_or_patterns_text is not None
        self._suffixes = pygount.common.regexes_from(regexes_or_patterns_text, _DEFAULT_SUFFIXES, source)

    def argument_parser(self):
        parser = argparse.ArgumentParser(description="count source lines of code", epilog=_HELP_EPILOG)
        parser.add_argument("--duplicates", "-d", action="store_true", help="analyze duplicate files")
        parser.add_argument("--encoding", "-e", default=_DEFAULT_ENCODING, help=_HELP_ENCODING)
        parser.add_argument(
            "--folders-to-skip",
            "-F",
            metavar="PATTERNS",
            default=pygount.analysis.DEFAULT_FOLDER_PATTERNS_TO_SKIP_TEXT,
            help=_HELP_FOLDERS_TO_SKIP,
        )
        parser.add_argument(
            "--format",
            "-f",
            metavar="FORMAT",
            choices=VALID_OUTPUT_FORMATS,
            default=_DEFAULT_OUTPUT_FORMAT,
            help=_HELP_FORMAT,
        )
        parser.add_argument(
            "--generated",
            "-g",
            metavar="PATTERNS",
            default=pygount.analysis.DEFAULT_GENERATED_PATTERNS_TEXT,
            help=_HELP_GENERATED,
        )
        parser.add_argument(
            "--merge-embedded-languages",
            "-m",
            action="store_true",
            help=_HELP_MERGE_EMBEDDED_LANGUAGES,
        )
        parser.add_argument(
            "--names-to-skip",
            "-N",
            metavar="PATTERNS",
            default=pygount.analysis.DEFAULT_NAME_PATTERNS_TO_SKIP_TEXT,
            help=_HELP_NAMES_TO_SKIP,
        )
        parser.add_argument(
            "--out",
            "-o",
            metavar="FILE",
            default=_DEFAULT_OUTPUT,
            help='file to write results to; use "STDOUT" for standard output; default: "%(default)s"',
        )
        parser.add_argument("--suffix", "-s", metavar="PATTERNS", default=_DEFAULT_SUFFIXES, help=_HELP_SUFFIX)
        parser.add_argument(
            "source_patterns",
            metavar="SHELL-PATTERN",
            nargs="*",
            default=[os.getcwd()],
            help="source files and directories to scan; can use glob patterns; default: current directory",
        )
        parser.add_argument("--verbose", "-v", action="store_true", help="explain what is being done")
        parser.add_argument("--version", action="version", version="%(prog)s " + pygount.__version__)
        return parser

    def parsed_args(self, arguments):
        assert arguments is not None

        parser = self.argument_parser()
        args = parser.parse_args(arguments)
        if args.encoding == "automatic":
            default_encoding = args.encoding
            fallback_encoding = None
        elif args.encoding == "chardet":
            if not pygount.analysis.has_chardet:  # pragma: no cover
                parser.error("chardet must be installed in order to specify --encoding=chardet")
            default_encoding = args.encoding
            fallback_encoding = None
        else:
            if args.encoding.startswith("automatic;"):
                first_encoding_semicolon_index = args.encoding.find(";")
                default_encoding = args.encoding[:first_encoding_semicolon_index]
                fallback_encoding = args.encoding[first_encoding_semicolon_index + 1 :]
                encoding_to_check = ("fallback encoding", fallback_encoding)
            else:
                default_encoding = args.encoding
                fallback_encoding = None
                encoding_to_check = ("encoding", default_encoding)
            if encoding_to_check is not None:
                name, encoding = encoding_to_check
                try:
                    "".encode(encoding)
                except LookupError:
                    parser.error(f"{name} specified with --encoding must be a known Python encoding: {encoding}")
        return args, default_encoding, fallback_encoding

    def apply_arguments(self, arguments=None):
        if arguments is None:  # pragma: no cover
            arguments = sys.argv[1:]
        args, default_encoding, fallback_encoding = self.parsed_args(arguments)
        self.set_default_encoding(default_encoding, "option --encoding")
        self.set_fallback_encoding(fallback_encoding, "option --encoding")
        self.set_folders_to_skip(args.folders_to_skip, "option --folders-to-skip")
        self.set_generated_regexps(args.generated, "option --generated")
        self.set_has_duplicates(args.duplicates, "option --duplicates")
        self.set_has_to_merge_embedded_languages(args.merge_embedded_languages, "option --merge-embedded-languages")
        self.set_is_verbose(args.verbose, "option --verbose")
        self.set_names_to_skip(args.names_to_skip, "option --folders-to-skip")
        self.set_output(args.out, "option --out")
        self.set_output_format(args.format, "option --format")
        self.set_source_patterns(args.source_patterns, "option PATTERNS")
        self.set_suffixes(args.suffix, "option --suffix")

    def execute(self):
        _log.setLevel(logging.INFO if self.is_verbose else logging.WARNING)
        with pygount.analysis.SourceScanner(
            self.source_patterns, self.suffixes, self.folders_to_skip, self.names_to_skip
        ) as source_scanner:
            source_paths_and_groups_to_analyze = list(source_scanner.source_paths())
            duplicate_pool = pygount.analysis.DuplicatePool() if not self.has_duplicates else None
            writer_class = _OUTPUT_FORMAT_TO_WRITER_CLASS_MAP[self.output_format]
            is_stdout = self.output == "STDOUT"
            target_context_manager = (
                contextlib.nullcontext(sys.stdout)
                if is_stdout
                else open(self.output, "w", encoding="utf-8", newline="")  # noqa: SIM115
            )
            with (
                target_context_manager as target_file,
                writer_class(target_file) as writer,
                Progress(disable=not writer.has_to_track_progress, transient=True) as progress,
            ):
                try:
                    for path_data in progress.track(source_paths_and_groups_to_analyze):
                        writer.add(
                            pygount.analysis.SourceAnalysis.from_file(
                                path_data.source_path,
                                path_data.group,
                                self.default_encoding,
                                self.fallback_encoding,
                                generated_regexes=self._generated_regexs,
                                duplicate_pool=duplicate_pool,
                                merge_embedded_language=self.has_to_merge_embedded_languages,
                                tmp_dir=path_data.tmp_dir,
                            )
                        )
                finally:
                    progress.stop()


def pygount_command(arguments=None):
    result = 1
    command = Command()
    try:
        command.apply_arguments(arguments)
        command.execute()
        result = 0
    except KeyboardInterrupt:  # pragma: no cover
        _log.error("interrupted as requested by user")
    except (pygount.common.OptionError, OSError) as error:
        _log.error(error)
    except Exception as error:
        _log.exception(error)

    return result


def main():  # pragma: no cover
    logging.basicConfig(level=logging.WARNING)
    sys.exit(pygount_command())


if __name__ == "__main__":  # pragma: no cover
    main()



"""
Summaries of analyses of multiple source codes.
"""

# Copyright (c) 2016-2024, Thomas Aglassinger.
# All rights reserved. Distributed under the BSD License.
import functools
import re
from collections.abc import Hashable

from .analysis import SourceAnalysis
from .common import mapped_repr

_PSEUDO_LANGUAGE_REGEX = re.compile("^__[a-z]+__$")


@functools.total_ordering
class LanguageSummary:
    """
    Summary of a source code counts from multiple files of the same language.
    """

    def __init__(self, language: str):
        self._language = language
        self._code_count = 0
        self._documentation_count = 0
        self._empty_count = 0
        self._file_count = 0
        self._file_percentage = 0.0
        self._string_count = 0
        self._is_pseudo_language = _PSEUDO_LANGUAGE_REGEX.match(self.language) is not None
        self._has_up_to_date_percentages = False

    @property
    def language(self) -> str:
        """the language to be summarized"""
        return self._language

    @property
    def code_count(self) -> int:
        """sum lines of code for this language"""
        return self._code_count

    @property
    def code_percentage(self) -> float:
        """percentage of lines containing code for this language across entire project"""
        return _percentage_or_0(self.code_count, self.line_count)

    def _assert_has_up_to_date_percentages(self):
        assert self._has_up_to_date_percentages, "update_percentages() must be called first"

    @property
    def documentation_count(self) -> int:
        """sum lines of documentation for this language"""
        return self._documentation_count

    @property
    def documentation_percentage(self) -> float:
        """percentage of lines containing documentation for this language across entire project"""
        return _percentage_or_0(self.documentation_count, self.line_count)

    @property
    def empty_count(self) -> int:
        """sum empty lines for this language"""
        return self._empty_count

    @property
    def empty_percentage(self) -> float:
        """percentage of empty lines for this language across entire project"""
        return _percentage_or_0(self.empty_count, self.line_count)

    @property
    def file_count(self) -> int:
        """number of source code files for this language"""
        return self._file_count

    @property
    def file_percentage(self) -> float:
        """percentage of files in project"""
        self._assert_has_up_to_date_percentages()
        return self._file_percentage

    @property
    def line_count(self) -> int:
        """sum count of all lines of any kind for this language"""
        return self.code_count + self.documentation_count + self.empty_count + self.string_count

    @property
    def string_count(self) -> int:
        """sum number of lines containing strings for this language"""
        return self._string_count

    @property
    def string_percentage(self) -> float:
        """percentage of lines containing strings for this language across entire project"""
        return _percentage_or_0(self.string_count, self.line_count)

    @property
    def source_count(self) -> int:
        """sum number of source lines of code"""
        return self.code_count + self.string_count

    @property
    def source_percentage(self) -> float:
        """percentage of source lines for code for this language across entire project"""
        return _percentage_or_0(self.source_count, self.line_count)

    @property
    def is_pseudo_language(self) -> bool:
        """``True`` if the language is not a real programming language"""
        return self._is_pseudo_language

    def sort_key(self) -> Hashable:
        """sort key to sort multiple languages by importance"""
        return self.code_count, self.documentation_count, self.string_count, self.empty_count, self.language

    def __eq__(self, other):
        return self.sort_key() == other.sort_key()

    def __lt__(self, other):
        return self.sort_key() < other.sort_key()

    def add(self, source_analysis: SourceAnalysis) -> None:
        """
        Add counts from ``source_analysis`` to total counts for this language.
        """
        assert source_analysis is not None
        assert source_analysis.language == self.language

        self._has_up_to_date_percentages = False
        self._file_count += 1
        if source_analysis.is_countable:
            self._code_count += source_analysis.code_count
            self._documentation_count += source_analysis.documentation_count
            self._empty_count += source_analysis.empty_count
            self._string_count += source_analysis.string_count

    def update_file_percentage(self, project_summary: "ProjectSummary"):
        self._file_percentage = _percentage_or_0(self.file_count, project_summary.total_file_count)
        self._has_up_to_date_percentages = True

    def __repr__(self):
        name_to_value_map = {
            "language": f"{self.language!r}",
            "file_count": self.file_count,
        }
        if not self.is_pseudo_language:
            name_to_value_map.update(
                {
                    "code_count": self.code_count,
                    "documentation_count": self.documentation_count,
                    "empty_count": self.empty_count,
                    "string_count": self.string_count,
                }
            )
        return mapped_repr(self, name_to_value_map)
        result = ", ".join(f"{name}={value}" for name, value in name_to_value_map.items())
        result = f"{self.__class__.__name__}({result})"
        return result


def _percentage_or_0(partial_count: int, total_count: int) -> float:
    assert partial_count >= 0
    assert total_count >= 0
    return 100 * partial_count / total_count if total_count != 0 else 0.0


class ProjectSummary:
    """
    Summary of source code counts for several languages and files.
    """

    def __init__(self):
        self._language_to_language_summary_map = {}
        self._total_code_count = 0
        self._total_documentation_count = 0
        self._total_empty_count = 0
        self._total_string_count = 0
        self._total_file_count = 0
        self._total_line_count = 0

    @property
    def language_to_language_summary_map(self) -> dict[str, LanguageSummary]:
        """
        A map containing summarized counts for each language added with :py:meth:`add()` so far.
        """
        return self._language_to_language_summary_map

    @property
    def total_code_count(self) -> int:
        return self._total_code_count

    @property
    def total_code_percentage(self) -> float:
        return _percentage_or_0(self.total_code_count, self.total_line_count)

    @property
    def total_documentation_count(self) -> int:
        return self._total_documentation_count

    @property
    def total_documentation_percentage(self) -> float:
        return _percentage_or_0(self.total_documentation_count, self.total_line_count)

    @property
    def total_empty_count(self) -> int:
        return self._total_empty_count

    @property
    def total_empty_percentage(self) -> float:
        return _percentage_or_0(self.total_empty_count, self.total_line_count)

    @property
    def total_file_count(self) -> int:
        return self._total_file_count

    @property
    def total_line_count(self) -> int:
        return self._total_line_count

    @property
    def total_source_count(self) -> int:
        return self.total_code_count + self.total_string_count

    @property
    def total_source_percentage(self) -> float:
        return _percentage_or_0(self.total_source_count, self.total_line_count)

    @property
    def total_string_count(self) -> int:
        return self._total_string_count

    @property
    def total_string_percentage(self) -> float:
        return _percentage_or_0(self.total_string_count, self.total_line_count)

    def add(self, source_analysis: SourceAnalysis) -> None:
        """
        Add counts from ``source_analysis`` to total counts.
        """
        self._total_file_count += 1
        language_summary = self.language_to_language_summary_map.get(source_analysis.language)
        if language_summary is None:
            language_summary = LanguageSummary(source_analysis.language)
            self.language_to_language_summary_map[source_analysis.language] = language_summary
        language_summary.add(source_analysis)

        if source_analysis.is_countable:
            self._total_code_count += source_analysis.code_count
            self._total_documentation_count += source_analysis.documentation_count
            self._total_empty_count += source_analysis.empty_count
            self._total_line_count += (
                source_analysis.code_count
                + source_analysis.documentation_count
                + source_analysis.empty_count
                + source_analysis.string_count
            )
            self._total_string_count += source_analysis.string_count

    def update_file_percentages(self) -> None:
        """Update percentages for all languages part of the project."""
        for language_summary in self._language_to_language_summary_map.values():
            language_summary.update_file_percentage(self)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"total_file_count={self.total_file_count}, "
            f"total_line_count={self.total_line_count}, "
            f"languages={sorted(self.language_to_language_summary_map.keys())})"
        )



"""
Common classes and functions for pygount.
"""

# Copyright (c) 2016-2024, Thomas Aglassinger.
# All rights reserved. Distributed under the BSD License.
import fnmatch
import functools
import inspect
import re
import warnings
from collections.abc import Iterator, Sequence
from re import Pattern
from typing import Optional, Union

#: Pseudo pattern to indicate that the remaining pattern are an addition to the default patterns.
ADDITIONAL_PATTERN = "[...]"

#: Prefix to use for pattern strings to describe a regular expression instead of a shell pattern.
REGEX_PATTERN_PREFIX = "[regex]"

_REGEX_TYPE = type(re.compile(""))


class Error(Exception):
    """
    Error to indicate that something went wrong during a pygount run.
    """


class OptionError(Error):
    """
    Error to indicate that a value passed to a command line option must be
    fixed.
    """

    def __init__(self, message, source=None):
        super().__init__(message)
        self.option_error_message = (source + ": ") if source is not None else ""
        self.option_error_message += message

    def __str__(self):
        return self.option_error_message


def as_list(items_or_text: Union[str, Sequence[str]]) -> list[str]:
    if isinstance(items_or_text, str):
        # TODO: Allow to specify comma (,) in text using '[,]'.
        result = [item.strip() for item in items_or_text.split(",") if item.strip() != ""]
    else:
        result = list(items_or_text)
    return result


def regex_from(pattern: Union[str, Pattern], is_shell_pattern=False) -> Pattern:
    assert pattern is not None
    if isinstance(pattern, str):
        result = re.compile(fnmatch.translate(pattern)) if is_shell_pattern else re.compile(pattern)
    else:
        result = pattern  # Assume pattern already is a compiled regular expression
    return result


def regexes_from(
    patterns_text: Union[str, Sequence[str], Sequence[Pattern]],
    default_patterns_text: Optional[Union[str, Sequence[Pattern], Sequence[str]]] = None,
    source: Optional[str] = None,
) -> list[Pattern]:
    assert patterns_text is not None

    result = []
    default_regexes = []
    try:
        if isinstance(patterns_text, str):
            is_shell_pattern = True
            patterns_text_without_prefixes = patterns_text
            if patterns_text_without_prefixes.startswith(REGEX_PATTERN_PREFIX):
                is_shell_pattern = False
                patterns_text_without_prefixes = patterns_text_without_prefixes[len(REGEX_PATTERN_PREFIX) :]
            if patterns_text_without_prefixes.startswith(ADDITIONAL_PATTERN):
                assert default_patterns_text is not None
                default_regexes = regexes_from(default_patterns_text)
                patterns_text_without_prefixes = patterns_text_without_prefixes[len(ADDITIONAL_PATTERN) :]

            patterns = as_list(patterns_text_without_prefixes)
            result = [regex_from(pattern, is_shell_pattern) for pattern in patterns]
        else:
            regexes = list(patterns_text)
            if len(regexes) >= 1 and regexes[0] is None:
                default_regexes = regexes_from(default_patterns_text)
                regexes = regexes[1:]
            for supposed_regex in regexes:
                assert isinstance(supposed_regex, _REGEX_TYPE), (
                    f"patterns_text must a text or sequence or regular expressions but contains: {supposed_regex}"
                )
            result.extend(regexes)
    except re.error as error:
        raise OptionError(f"cannot parse pattern for regular repression: {error}", source) from None
    result.extend(default_regexes)
    return result


def lines(text: str) -> Iterator[str]:
    """
    Generator function to yield lines (delimited with ``'\n'``) stored in
    ``text``. This is useful when a regular expression should only match on a
    per-line basis in a memory efficient way.
    """
    assert text is not None
    assert "\r" not in text
    previous_newline_index = 0
    newline_index = text.find("\n")
    while newline_index != -1:
        yield text[previous_newline_index:newline_index]
        previous_newline_index = newline_index + 1
        newline_index = text.find("\n", previous_newline_index)
    last_line = text[previous_newline_index:]
    if last_line != "":
        yield last_line


def deprecated(reason: Optional[str]):  # pragma: no cover
    """
    Decorator to mark functions as deprecated and log a warning in case it is called.

    Source: https://stackoverflow.com/questions/2536307/decorators-in-the-python-standard-lib-deprecated-specifically
    """

    if isinstance(reason, str):
        # The @deprecated is used with a 'reason'.
        #
        # .. code-block:: python
        #
        #    @deprecated("please, use another function")
        #    def old_function(x, y):
        #      pass

        def decorator(func1):
            if inspect.isclass(func1):
                fmt1 = "Call to deprecated class {name} ({reason})."
            else:
                fmt1 = "Call to deprecated function {name} ({reason})."

            @functools.wraps(func1)
            def new_func1(*args, **kwargs):
                warnings.simplefilter("always", DeprecationWarning)
                warnings.warn(
                    fmt1.format(name=func1.__name__, reason=reason), category=DeprecationWarning, stacklevel=2
                )
                warnings.simplefilter("default", DeprecationWarning)
                return func1(*args, **kwargs)

            return new_func1

        return decorator
    if inspect.isclass(reason) or inspect.isfunction(reason):
        # The @deprecated is used without any 'reason'.
        #
        # .. code-block:: python
        #
        #    @deprecated
        #    def old_function(x, y):
        #      pass

        func2 = reason
        fmt2 = "Call to deprecated class {name}." if inspect.isclass(func2) else "Call to deprecated function {name}."

        @functools.wraps(func2)
        def new_func2(*args, **kwargs):
            warnings.simplefilter("always", DeprecationWarning)
            warnings.warn(fmt2.format(name=func2.__name__), category=DeprecationWarning, stacklevel=2)
            warnings.simplefilter("default", DeprecationWarning)
            return func2(*args, **kwargs)

        return new_func2
    raise TypeError(repr(type(reason)))


def mapped_repr(type_, name_to_value_map) -> str:
    result = ", ".join(f"{name}={value}" for name, value in name_to_value_map.items())
    result = f"{type_.__class__.__name__}({result})"
    return result



"""
Pygount counts lines of source code using pygments lexers.
"""

# Copyright (c) 2016-2024, Thomas Aglassinger.
# All rights reserved. Distributed under the BSD License.
from importlib.metadata import version

from .analysis import DuplicatePool, SourceAnalysis, SourceScanner, SourceState, encoding_for
from .common import Error, OptionError
from .summary import LanguageSummary, ProjectSummary

__version__ = version(__name__)

__all__ = [
    "DuplicatePool",
    "Error",
    "LanguageSummary",
    "OptionError",
    "ProjectSummary",
    "SourceAnalysis",
    "SourceScanner",
    "SourceState",
    "__version__",
    "encoding_for",
]



"""
Additional lexers for pygount that fill gaps left by :py:mod:`pygments`.
"""

# Copyright (c) 2016-2024, Thomas Aglassinger.
# All rights reserved. Distributed under the BSD License.
import pygments.lexer
import pygments.lexers
import pygments.token
import pygments.util


class IdlLexer(pygments.lexers.JavaLexer):
    """
    Lexer for OMG Interface Definition Language (IDL) that simply uses the
    existing Java lexer to find comments. While this is useless for syntax
    highlighting it is good enough for counting lines.
    """

    name = "IDL"
    filenames = ["*.idl"]


class MinimalisticM4Lexer(pygments.lexer.RegexLexer):
    """
    Minimalistic lexer for m4 macro processor that can distinguish between
    comments and code. It does not recognize a redefined comment mark though.
    """

    name = "M4"
    tokens = {
        "root": [
            (r"(.*)(#.*\n)", pygments.lexer.bygroups(pygments.token.Text, pygments.token.Comment.Single)),
            (r".*\n", pygments.token.Text),
        ]
    }


class MinimalisticVBScriptLexer(pygments.lexer.RegexLexer):
    """
    Minimalistic lexer for VBScript that can distinguish between comments and
    code.
    """

    name = "VBScript"
    tokens = {"root": [(r"\s*'.*\n", pygments.token.Comment.Single), (r".*\n", pygments.token.Text)]}


class MinimalisticWebFocusLexer(pygments.lexer.RegexLexer):
    """
    Minimalistic lexer for WebFOCUS that can distinguish between comments and
    code.
    """

    name = "WebFOCUS"
    tokens = {"root": [(r"-\*.*\n", pygments.token.Comment.Single), (r".*\n", pygments.token.Text)]}


class PlainTextLexer(pygments.lexer.RegexLexer):
    """
    Simple lexer for plain text that treats every line with non-white space
    characters as :py:data:`pygments.Token.Comment.Single` and only lines
    that are empty or contain only white space as
    :py:data:`pygments.Token.Text`.

    This way, plaint text files count as documentation.
    """

    name = "Text"
    tokens = {"root": [(r"\s*\n", pygments.token.Text), (r".+\n", pygments.token.Comment.Single)]}



"""
Function to obtain the language dialect used by XML source code.
"""

# Copyright (c) 2016-2024, Thomas Aglassinger.
# All rights reserved. Distributed under the BSD License.
import logging
import re
import xml.sax

# TODO #10: Replace regex for DTD by working DTD handler.
#: Regular expression to obtain DTD.
_DTD_REGEX = re.compile(r'<!DOCTYPE\s+(?P<name>[a-zA-Z][a-zA-Z-]*)\s+PUBLIC\s+"(?P<public_id>.+)"')
_REGEX_PATTERNS_AND_DIALECTS = ((".*DocBook.*", "DocBook XML"),)
_REGEXES_AND_DIALECTS = [(re.compile(pattern), dialect) for pattern, dialect in _REGEX_PATTERNS_AND_DIALECTS]
for public_id_regex, dialect in _REGEX_PATTERNS_AND_DIALECTS:
    assert public_id_regex is not None
    assert dialect is not None
    assert dialect.strip() != ""
#: Regex to detect Sax error messages with uninformative paths like '<unknown>'.
_SAX_MESSAGE_WITHOUT_PATH_PATTERN = re.compile(r"^<.+>(?P<message_without_path>:\d+:\d+.+)")

_log = logging.getLogger("pygount")


class SaxParserDone(Exception):
    """
    Pseudo error to indicate that the Sax parser ist done.
    """


class XmlDialectHandler(xml.sax.ContentHandler, xml.sax.handler.DTDHandler):
    def __init__(self, max_element_count=100):
        super().__init__()
        self.dialect = None
        self._path = ""
        self._element_count = 0
        self._max_element_count = max_element_count

    def _set_dialect_and_stop_parsing(self, dialect):
        self.dialect = dialect
        raise SaxParserDone(f"language detected: {dialect}")

    def startElement(self, name, attrs):
        self._element_count += 1
        if self._element_count == self._max_element_count:
            raise SaxParserDone(f"no language found after parsing {self._element_count} elements")
        self._path += "/" + name
        xmlns = attrs.get("xmlns", "")
        if (self._path == "/project") and ("name" in attrs):
            self._set_dialect_and_stop_parsing("Ant")
        elif (self._path in ("/book/title", "/chapter/title")) or (xmlns == "http://docbook.org/ns/docbook"):
            self._set_dialect_and_stop_parsing("DocBook XML")
        elif xmlns == "http://xmlns.jcp.org/xml/ns/javaee":
            self._set_dialect_and_stop_parsing("JavaEE XML")
        elif xmlns.startswith("http://maven.apache.org/POM"):
            self._set_dialect_and_stop_parsing("Maven")
        elif xmlns.startswith("http://www.netbeans.org/ns/project/"):
            self._set_dialect_and_stop_parsing("NetBeans Project")

    def endElement(self, name):
        self._path = self._path[: -len(name) - 1]


def xml_dialect(xml_path, xml_code):
    # TODO #10: Remove hack to obtain DTD using a regex instead of a DTDHandler.
    dtd_match = _DTD_REGEX.match(xml_code)
    if dtd_match is not None:
        public_id = dtd_match.group("public_id")
        for public_id_regex, dialect in _REGEXES_AND_DIALECTS:
            if public_id_regex.match(public_id):
                return dialect

    xml_dialect_handler = XmlDialectHandler()
    parser = xml.sax.make_parser()
    parser.setContentHandler(xml_dialect_handler)
    parser.setFeature(xml.sax.handler.feature_external_ges, False)
    parser.setFeature(xml.sax.handler.feature_external_pes, False)
    parser.setFeature(xml.sax.handler.feature_validation, False)
    try:
        parser.feed(xml_code)
        # NOTE: We can only call close() when the parser has finished,
        # otherwise close() raises a SAXException('parser finished').
        parser.close()
    except SaxParserDone:
        # Language has been determined or the parser has given up.
        pass
    except (ValueError, xml.sax.SAXException) as error:
        # NOTE: ValueError is raised on unknown url type.
        error_message = str(error)
        message_without_path_match = _SAX_MESSAGE_WITHOUT_PATH_PATTERN.match(error_message)
        if message_without_path_match is not None:
            # HACK: Replace uninformative sax path like '<unknown>' with actual XML path.
            error_message = xml_path + message_without_path_match.group("message_without_path")
        _log.warning(error_message)
    except OSError as error:
        _log.warning("%s: cannot analyze XML dialect: %s", xml_path, error)
    return xml_dialect_handler.dialect



"""
Writers to store the results of a pygount analysis.
"""

# Copyright (c) 2016-2024, Thomas Aglassinger.
# All rights reserved. Distributed under the BSD License.
import datetime
import json
import math
import os
from xml.etree import ElementTree

from rich.console import Console
from rich.table import Table

import pygount

from . import SourceAnalysis
from .summary import ProjectSummary

#: Version of cloc the --format=cloc-xml pretends to be.
CLOC_VERSION = "1.60"

JSON_FORMAT_VERSION = "1.1.0"


class BaseWriter:
    def __init__(self, target_stream):
        self._target_stream = target_stream
        try:
            self.target_name = self._target_stream.name
        except AttributeError:
            self.target_name = "<io>"
        self.project_summary = ProjectSummary()
        self.started_at = self._utc_now()
        self.finished_at = None
        self.files_per_second = 0
        self.lines_per_second = 0
        self.duration = None
        self.duration_in_seconds = 0.0
        self.has_to_track_progress = True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    def add(self, source_analysis):
        self.project_summary.add(source_analysis)

    def close(self):
        self.project_summary.update_file_percentages()
        self.finished_at = self._utc_now()
        self.duration = self.finished_at - self.started_at
        self.duration_in_seconds = max(
            0.001, self.duration.microseconds * 1e-6 + self.duration.seconds + self.duration.days * 3600 * 24
        )
        self.lines_per_second = self.project_summary.total_line_count / self.duration_in_seconds
        self.files_per_second = self.project_summary.total_file_count / self.duration_in_seconds

    @staticmethod
    def _utc_now() -> datetime.datetime:
        # After switching to Python 3.11+, we can change this to `now(datetime.UTC)`.
        return datetime.datetime.now(datetime.timezone.utc)


class LineWriter(BaseWriter):
    """
    Writer that simply writes a line of text for each source code.
    """

    def __init__(self, target_stream):
        super().__init__(target_stream)
        self.has_to_track_progress = False
        self.template = "{0}\t{1}\t{2}\t{3}"

    def add(self, source_analysis):
        source_line_count = source_analysis.code_count + source_analysis.string_count
        line_to_write = self.template.format(
            source_line_count, source_analysis.language, source_analysis.group, source_analysis.path
        )
        self._target_stream.write(line_to_write + os.linesep)


class ClocXmlWriter(BaseWriter):
    """
    Writer that writes XML output similar to cloc when called with options
    --by-file --xml. This kind of output can be processed by Jenkins' SLOCCount
    plug-in.
    """

    def __init__(self, target_stream):
        super().__init__(target_stream)
        self._results_element = ElementTree.Element("results")
        self._header_element = ElementTree.SubElement(self._results_element, "header")
        ElementTree.SubElement(self._header_element, "cloc_url", text="https://github.com/roskakori/pygount")
        ElementTree.SubElement(self._header_element, "cloc_version", text=CLOC_VERSION)
        self._files_element = ElementTree.SubElement(self._results_element, "files")

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            # Only write the XML if everything works out.
            self.close()

    def add(self, source_analysis: SourceAnalysis):
        super().add(source_analysis)
        file_attributes = {
            "blank": str(source_analysis.empty_count),
            "code": str(source_analysis.source_count),
            "comment": str(source_analysis.documentation_count),
            "language": source_analysis.language,
            "name": source_analysis.path,
        }
        ElementTree.SubElement(self._files_element, "file", attrib=file_attributes)

    def close(self):
        super().close()
        # Add various statistics to <header>.
        ElementTree.SubElement(self._header_element, "elapsed_seconds", text=str(self.duration_in_seconds))
        ElementTree.SubElement(self._header_element, "n_files", text=str(self.project_summary.total_file_count))
        ElementTree.SubElement(self._header_element, "n_lines", text=str(self.project_summary.total_line_count))
        ElementTree.SubElement(self._header_element, "files_per_second", text=f"{self.files_per_second:f}")
        ElementTree.SubElement(self._header_element, "lines_per_second", text=f"{self.lines_per_second:f}")
        ElementTree.SubElement(self._header_element, "report_file", text=self.target_name)

        # Add totals to <files>.
        file_attributes = {
            "blank": str(self.project_summary.total_empty_count),
            "code": str(self.project_summary.total_code_count + self.project_summary.total_string_count),
            "comment": str(self.project_summary.total_documentation_count),
        }
        ElementTree.SubElement(self._files_element, "total", attrib=file_attributes)

        # Write the whole XML file.
        if self._target_stream.encoding is not None:
            # Write XML declaration only for files but skip it for io.StringIO.
            self._target_stream.write(f'<?xml version="1.0" encoding="{self._target_stream.encoding}"?>')
        xml_root = ElementTree.ElementTree(self._results_element)
        xml_root.write(self._target_stream, encoding="unicode", xml_declaration=False)


class SummaryWriter(BaseWriter):
    """
    Writer to summarize the analysis per language in a format that can easily
    be read by humans.
    """

    _COLUMNS_WITH_JUSTIFY = (
        ("Language", "left"),
        ("Files", "right"),
        ("%", "right"),
        ("Code", "right"),
        ("%", "right"),
        ("Comment", "right"),
        ("%", "right"),
    )

    def close(self):
        super().close()

        table = Table()
        for column, justify in self._COLUMNS_WITH_JUSTIFY:
            table.add_column(column, justify=justify, overflow="fold")

        language_summaries = sorted(self.project_summary.language_to_language_summary_map.values(), reverse=True)
        for index, language_summary in enumerate(language_summaries, start=1):
            table.add_row(
                language_summary.language,
                str(language_summary.file_count),
                formatted_percentage(language_summary.file_percentage),
                str(language_summary.code_count),
                formatted_percentage(language_summary.code_percentage),
                str(language_summary.documentation_count),
                formatted_percentage(language_summary.documentation_percentage),
                end_section=(index == len(language_summaries)),
            )
        table.add_row(
            "Sum",
            str(self.project_summary.total_file_count),
            formatted_percentage(100.0),
            str(self.project_summary.total_code_count),
            formatted_percentage(self.project_summary.total_code_percentage),
            str(self.project_summary.total_documentation_count),
            formatted_percentage(self.project_summary.total_documentation_percentage),
        )
        Console(file=self._target_stream, soft_wrap=True).print(table)


class JsonWriter(BaseWriter):
    """
    Writer JSON output, ideal for further automatic processing.
    """

    def __init__(self, target_stream):
        super().__init__(target_stream)
        self.source_analyses = []

    def add(self, source_analysis: SourceAnalysis):
        super().add(source_analysis)
        self.source_analyses.append(
            {
                "codeCount": source_analysis.code_count,
                "documentationCount": source_analysis.documentation_count,
                "emptyCount": source_analysis.empty_count,
                "group": source_analysis.group,
                "isCountable": source_analysis.is_countable,
                "language": source_analysis.language,
                "lineCount": source_analysis.line_count,
                "path": source_analysis.path,
                "state": source_analysis.state.name,
                "stateInfo": source_analysis.state_info,
                "sourceCount": source_analysis.source_count,
            }
        )

    def close(self):
        # NOTE: JSON names use camel case to follow JSLint's guidelines, see <https://www.jslint.com/>.
        super().close()
        json_map = {
            "formatVersion": JSON_FORMAT_VERSION,
            "pygountVersion": pygount.__version__,
            "files": self.source_analyses,
            "languages": [
                {
                    "documentationCount": language_summary.documentation_count,
                    "documentationPercentage": language_summary.documentation_percentage,
                    "codeCount": language_summary.code_count,
                    "codePercentage": language_summary.code_percentage,
                    "emptyCount": language_summary.empty_count,
                    "emptyPercentage": language_summary.empty_percentage,
                    "fileCount": language_summary.file_count,
                    "filePercentage": language_summary.file_percentage,
                    "isPseudoLanguage": language_summary.is_pseudo_language,
                    "language": language_summary.language,
                    "sourceCount": language_summary.source_count,
                    "sourcePercentage": language_summary.source_percentage,
                    "stringCount": language_summary.string_count,
                    "stringPercentage": language_summary.string_percentage,
                }
                for language_summary in self.project_summary.language_to_language_summary_map.values()
            ],
            "runtime": {
                "durationInSeconds": self.duration_in_seconds,
                "filesPerSecond": self.files_per_second,
                "finishedAt": self.finished_at.isoformat(),
                "linesPerSecond": self.lines_per_second,
                "startedAt": self.started_at.isoformat(),
            },
            "summary": {
                "totalCodeCount": self.project_summary.total_code_count,
                "totalCodePercentage": self.project_summary.total_code_percentage,
                "totalDocumentationCount": self.project_summary.total_documentation_count,
                "totalDocumentationPercentage": self.project_summary.total_documentation_percentage,
                "totalEmptyCount": self.project_summary.total_empty_count,
                "totalEmptyPercentage": self.project_summary.total_empty_percentage,
                "totalFileCount": self.project_summary.total_file_count,
                "totalSourceCount": self.project_summary.total_source_count,
                "totalSourcePercentage": self.project_summary.total_source_percentage,
                "totalStringCount": self.project_summary.total_string_count,
                "totalStringPercentage": self.project_summary.total_string_percentage,
            },
        }
        json.dump(json_map, self._target_stream)


def digit_width(line_count: int) -> int:
    assert line_count >= 0
    return math.ceil(math.log10(line_count + 1)) if line_count != 0 else 1


def formatted_percentage(percentage: float) -> str:
    assert percentage >= 0.0
    assert percentage <= 100.0
    return f"{percentage:.01f}"



import re
import shutil
from tempfile import mkdtemp
from typing import Optional

import git

#: Regular expression to detect git url with the optional tag or branch
# from https://stackoverflow.com/questions/2514859/regular-expression-for-git-repository server-name
_GIT_URL_REGEX = re.compile(
    r"(?P<remote_url>((git|ssh|http(s)?)|(git@[\w.-]+))(:(//)?)([\w.@:/\-~]+)(\.git))(/)?(?P<revision>[\w./\-]+)?"
)


def git_remote_url_and_revision_if_any(git_url: str) -> tuple[Optional[str], Optional[str]]:
    assert git_url is not None
    git_url_match = _GIT_URL_REGEX.match(git_url)
    return (
        (None, None) if git_url_match is None else (git_url_match.group("remote_url"), git_url_match.group("revision"))
    )


class GitStorage:
    def __init__(self, remote_url: str, revision: Optional[str] = None):
        assert remote_url is not None
        self._remote_url = remote_url
        self._revision = revision
        self._temp_folder = mkdtemp()

    @property
    def temp_folder(self) -> str:
        return self._temp_folder

    def extract(self):
        multi_options = ["--depth", "1"]
        if self._revision is not None:
            multi_options.extend(["--branch", self._revision])
        git.Repo.clone_from(self._remote_url, self._temp_folder, multi_options=multi_options)

    def close(self):
        shutil.rmtree(self._temp_folder, ignore_errors=True)



"""
Functions to analyze source code and count lines in it.
"""

# Copyright (c) 2016-2024, Thomas Aglassinger.
# All rights reserved. Distributed under the BSD License.
import codecs
import collections
import glob
import hashlib
import itertools
import logging
import os
import re
from collections.abc import Iterator, Sequence
from dataclasses import dataclass
from enum import Enum
from io import SEEK_CUR, BufferedIOBase, IOBase, RawIOBase, TextIOBase
from re import Pattern
from typing import Optional, Union

import pygments.lexer
import pygments.lexers
import pygments.token
import pygments.util

import pygount.common
import pygount.lexers
import pygount.xmldialect
from pygount.common import mapped_repr
from pygount.git_storage import GitStorage, git_remote_url_and_revision_if_any

HTTP_URL_REGEX = re.compile(r"^(https?://)")
_ALLOWED_GIT_PLATFORMS = ["github.com", "bitbucket.org", "gitlab.com"]
_ALLOWED_GIT_PLATFORM_CHOICES_PATTERN = "|".join(map(re.escape, _ALLOWED_GIT_PLATFORMS))
GIT_REPO_REGEX = re.compile(rf"^(https?://|git@)({_ALLOWED_GIT_PLATFORM_CHOICES_PATTERN})/[^/]+/[^/]+")

# Attempt to import chardet.
try:
    import chardet.universaldetector

    _detector = chardet.universaldetector.UniversalDetector()
except ImportError:
    _detector = None
has_chardet = bool(_detector)

#: Fallback encoding to use if no encoding is specified
DEFAULT_FALLBACK_ENCODING = "cp1252"

#: Default glob patterns for folders not to analyze.
DEFAULT_FOLDER_PATTERNS_TO_SKIP_TEXT = ", ".join(
    [".?*", "_svn", "__pycache__"]  # Subversion hack for Windows  # Python byte code
)

#: Pygments token type; we need to define our own type because pygments' ``_TokenType`` is internal.
TokenType = type(pygments.token.Token)

_BASE_LANGUAGE_REGEX = re.compile(r"^(?P<base_language>[^+]+)\+[^+].*$")

#: BOMs to indicate that a file is a text file even if it contains zero bytes.
_TEXT_BOMS = (codecs.BOM_UTF16_BE, codecs.BOM_UTF16_LE, codecs.BOM_UTF32_BE, codecs.BOM_UTF32_LE, codecs.BOM_UTF8)


class SourceState(Enum):
    """
    Possible values for :py:attr:`SourceAnalysis.state`.
    """

    #: successfully analyzed
    analyzed = 1
    #: source code is a binary
    binary = 2
    #: source code is an identical copy of another
    duplicate = 3
    #: source code is empty (file size = 0)
    empty = 4
    #: source could not be parsed
    error = 5
    #: source code has been generated
    generated = 6
    # TODO: 'huge' = auto()  # source code exceeds size limit
    #: pygments does not offer any lexer to analyze the source
    unknown = 7


#: Default patterns for regular expressions to detect generated code.
#: The '(?i)' indicates that the patterns are case-insensitive.
DEFAULT_GENERATED_PATTERNS_TEXT = pygount.common.REGEX_PATTERN_PREFIX + ", ".join(
    [
        r"(?i).*automatically generated",
        r"(?i).*do not edit",
        r"(?i).*generated with the .+ utility",
        r"(?i).*this is a generated file",
        r"(?i).*generated automatically",
    ]
)

#: Default glob patterns for file names not to analyze.
DEFAULT_NAME_PATTERNS_TO_SKIP_TEXT = ", ".join([".*", "*~"])

_log = logging.getLogger("pygount")

_MARK_TO_NAME_MAP = (("c", "code"), ("d", "documentation"), ("e", "empty"), ("s", "string"))
_BOM_TO_ENCODING_MAP = collections.OrderedDict(
    (
        # NOTE: We need an ordered dict due to the overlap between utf-32-le and utf-16-be.
        (codecs.BOM_UTF8, "utf-8-sig"),
        (codecs.BOM_UTF32_LE, "utf-32-le"),
        (codecs.BOM_UTF16_BE, "utf-16-be"),
        (codecs.BOM_UTF16_LE, "utf-16-le"),
        (codecs.BOM_UTF32_BE, "utf-32-be"),
    )
)
_XML_PROLOG_REGEX = re.compile(r'<\?xml\s+.*encoding="(?P<encoding>[-_.a-zA-Z0-9]+)".*\?>')
_MAGIC_COMMENT_LINE_START_REGEXES = [
    re.compile(f"^{pattern}\\s*(?P<remainder>.+)$", re.IGNORECASE)
    for pattern in [
        r"#+",  # Python, Ruby
        r"//+",  # C++, Dart, Java, ...
        r"/\*+",  # C etc
        r"--+",  # Ada, SQL, VHDL
        r";+",  # Assembly
        r"%+",  # Latex, MatLab, Prolog
        r"rem\s",  # Basic, Windows batch
        r"\*+",  # Pascal
        r"\{",  # Pascal
    ]
]
_MAGIC_COMMENT_LINE_REMAINDER_REGEXES = [
    re.compile(pattern, re.IGNORECASE)
    for pattern in [
        # Covers for example "encoding: cp1252" and "encoding=cp1252".
        r"(en)?coding\s*[:=]\s*(?P<encoding>[-_.a-z0-9]+)\b",
        # Covers for example "-*- coding: cp1252 -*-".
        r"-\*-\s*coding\s*[:=]\s*(?P<encoding>[-_.a-z0-9]+)\s*(;.+\s*)?-\*-\s*",
    ]
]

_STANDARD_PLAIN_TEXT_NAME_PATTERNS = (
    # Text files for (moribund) gnits standards.
    "authors",
    "bugs",
    "changelog",
    "copying",
    "install",
    "license",
    "news",
    "readme",
    "thanks",
    # GitHub community recommendations, see
    # <https://docs.github.com/en/communities/setting-up-your-project-for-healthy-contributions>.
    # By now, in practice most projects use a suffix like "*.md" but some older ones
    # still might have such files without suffix.
    "code_of_conduct",
    "contributing",
    "support",
    # Other common text files.
    "changes",
    "faq",
    "readme\\.1st",
    "read\\.me",
    "todo",
)
_PLAIN_TEXT_PATTERN = "(^" + "$)|(^".join(_STANDARD_PLAIN_TEXT_NAME_PATTERNS) + "$)"
#: Regular expression to detect plain text files by name.
_PLAIN_TEXT_NAME_REGEX = re.compile(_PLAIN_TEXT_PATTERN, re.IGNORECASE)

_MARK_UP_NAME_PATTERN = r"^.*\.(md|rst|txt|\d+)$"
_MARK_UP_NAME_REGEX = re.compile(_MARK_UP_NAME_PATTERN, re.IGNORECASE)

#: Mapping for file suffixes to lexers for which pygments offers no official one.
_SUFFIX_TO_FALLBACK_LEXER_MAP = {
    "fex": pygount.lexers.MinimalisticWebFocusLexer(),
    "idl": pygount.lexers.IdlLexer(),
    "m4": pygount.lexers.MinimalisticM4Lexer(),
    "txt": pygount.lexers.PlainTextLexer(),
    "vbe": pygount.lexers.MinimalisticVBScriptLexer(),
    "vbs": pygount.lexers.MinimalisticVBScriptLexer(),
}
for _oracle_suffix in ("pck", "pkb", "pks", "pls"):
    _SUFFIX_TO_FALLBACK_LEXER_MAP[_oracle_suffix] = pygments.lexers.get_lexer_by_name("plpgsql")


@dataclass(frozen=True)
class PathData:
    source_path: str
    group: str
    tmp_dir: Optional[str] = None


def is_markup_file(source_path: str) -> bool:
    return _MARK_UP_NAME_REGEX.match(os.path.basename(source_path)) is not None


class DuplicatePool:
    """
    A pool that collects information about potential duplicate files.
    """

    def __init__(self):
        self._size_to_paths_map = {}
        self._size_and_hash_to_path_map = {}

    @staticmethod
    def _hash_for(path_to_hash):
        buffer_size = 1024 * 1024
        sha256_hash = hashlib.sha256()
        with open(path_to_hash, "rb", buffer_size) as file_to_hash:
            data = file_to_hash.read(buffer_size)
            while len(data) >= 1:
                sha256_hash.update(data)
                data = file_to_hash.read(buffer_size)
        return sha256_hash.digest()

    def duplicate_path(self, source_path: str) -> Optional[str]:
        """
        Path to a duplicate for ``source_path`` or ``None`` if no duplicate exists.

        Internally information is stored to identify possible future duplicates of
        ``source_path``.
        """
        result = None
        source_size = os.path.getsize(source_path)
        paths_with_same_size = self._size_to_paths_map.get(source_size)
        if paths_with_same_size is None:
            self._size_to_paths_map[source_size] = [source_path]
        else:
            source_hash = DuplicatePool._hash_for(source_path)
            if len(paths_with_same_size) == 1:
                # Retrofit the initial path with the same size and its hash.
                initial_path_with_same_size = paths_with_same_size[0]
                initial_hash = DuplicatePool._hash_for(initial_path_with_same_size)
                self._size_and_hash_to_path_map[(source_size, initial_hash)] = initial_path_with_same_size
            result = self._size_and_hash_to_path_map.get((source_size, source_hash))
            self._size_and_hash_to_path_map[(source_size, source_hash)] = source_path
        return result


class SourceAnalysis:
    """
    Results from analyzing a source path.

    Prefer the factory methods :py:meth:`from_file()` and :py:meth:`from_state` to
    calling the constructor.
    """

    def __init__(
        self,
        path: str,
        language: str,
        group: str,
        code: int,
        documentation: int,
        empty: int,
        string: int,
        state: SourceState,
        state_info: Optional[str] = None,
    ):
        SourceAnalysis._check_state_info(state, state_info)
        self._path = path
        self._language = language
        self._group = group
        self._code = code
        self._documentation = documentation
        self._empty = empty
        self._string = string
        self._state = state
        self._state_info = state_info

    @staticmethod
    def from_state(
        source_path: str,
        group: str,
        state: SourceState,
        state_info: Optional[str] = None,
        tmp_dir: Optional[str] = None,
    ) -> "SourceAnalysis":
        """
        Factory method to create a :py:class:`SourceAnalysis` with all counts
        set to 0 and everything else according to the specified parameters.
        """
        assert source_path is not None
        assert group is not None
        assert state != SourceState.analyzed, "use from() for analyzable sources"
        SourceAnalysis._check_state_info(state, state_info)
        reduced_path = source_path.split(tmp_dir)[-1].lstrip(os.sep) if tmp_dir else source_path
        return SourceAnalysis(
            path=reduced_path,
            language=f"__{state.name}__",
            group=group,
            code=0,
            documentation=0,
            empty=0,
            string=0,
            state=state,
            state_info=state_info,
        )

    @staticmethod
    def _check_state_info(state: SourceState, state_info: Optional[str]):
        assert state_info is None or isinstance(state_info, str), (
            f"state_info must be be None or str but is: {state_info!r}"
        )

        states_that_require_state_info = [SourceState.duplicate, SourceState.error, SourceState.generated]
        assert (state in states_that_require_state_info) == (state_info is not None), (
            f"state={state} and state_info={state_info} "
            f"but state_info must be specified for the following states: {states_that_require_state_info}"
        )

    @staticmethod
    def from_file(
        source_path: str,
        group: str,
        encoding: str = "automatic",
        fallback_encoding: str = "cp1252",
        generated_regexes: Optional[list[Pattern]] = None,
        duplicate_pool: Optional[DuplicatePool] = None,
        file_handle: Optional[IOBase] = None,
        merge_embedded_language: bool = False,
        tmp_dir: Optional[str] = None,
    ) -> "SourceAnalysis":
        """
        Factory method to create a :py:class:`SourceAnalysis` by analyzing
        the source code in ``source_path`` or the open file ``file_handle``.

        :param source_path: path to source code to analyze
        :param group: name of a logical group the source code belongs to, e.g. a
          package.
        :param encoding: encoding according to :func:`encoding_for`
        :param fallback_encoding: fallback encoding according to
          :func:`encoding_for`
        :param generated_regexes: list of regular expression that if found within the first few lines
          if a source code identify is as generated source code for which SLOC should not be counted
        :param duplicate_pool: a :class:`DuplicatePool` where information about possible duplicates is
          collected, or ``None`` if possible duplicates should be counted multiple times.
        :param file_handle: a file-like object, or ``None`` to read and open the file from
          ``source_path``. If the file is open in text mode, it must be opened with the correct
          encoding.
        :param merge_embedded_language: If pygments detects a base and embedded language, the source
          code counts towards the base language. For example: "JavaScript+Lasso" counts as
          "JavaScript".
        :param tmp_dir: If a temporary directory was created, strip it from the path name. This happens
          right now only for git repositories.
        """
        assert encoding is not None

        result = None
        lexer = None
        source_code = None
        if file_handle is None:
            source_size = os.path.getsize(source_path)
            if source_size == 0:
                _log.info("%s: is empty", source_path)
                result = SourceAnalysis.from_state(source_path, group, SourceState.empty)
            elif is_binary_file(source_path):
                _log.info("%s: is binary", source_path)
                result = SourceAnalysis.from_state(source_path, group, SourceState.binary)
            elif not has_lexer(source_path):
                _log.info("%s: unknown language", source_path)
                result = SourceAnalysis.from_state(source_path, group, SourceState.unknown)
        if duplicate_pool is not None:
            duplicate_path = duplicate_pool.duplicate_path(source_path)
            if duplicate_path is not None:
                _log.info("%s: is a duplicate of %s", source_path, duplicate_path)
                result = SourceAnalysis.from_state(source_path, group, SourceState.duplicate, duplicate_path)
        if result is None:
            try:
                if file_handle is None:
                    if encoding in ("automatic", "chardet"):
                        encoding = encoding_for(source_path, encoding, fallback_encoding)
                    with open(source_path, encoding=encoding) as source_file:
                        source_code = source_file.read()
                elif not isinstance(file_handle, TextIOBase):
                    if encoding in ("automatic", "chardet"):
                        encoding = encoding_for(source_path, encoding, fallback_encoding, file_handle=file_handle)
                    source_code = file_handle.read().decode(encoding)
                else:
                    source_code = file_handle.read()
            except (LookupError, OSError, UnicodeError) as error:
                _log.warning("cannot read %s using encoding %s: %s", source_path, encoding, error)
                result = SourceAnalysis.from_state(source_path, group, SourceState.error, str(error))
            if result is None:
                lexer = guess_lexer(source_path, source_code)
                assert lexer is not None
        actual_generated_regexes = (
            generated_regexes
            if generated_regexes is not None
            else pygount.common.regexes_from(DEFAULT_GENERATED_PATTERNS_TEXT)
        )
        if (result is None) and (len(actual_generated_regexes) != 0):
            number_line_and_regex = matching_number_line_and_regex(
                pygount.common.lines(source_code), actual_generated_regexes
            )
            if number_line_and_regex is not None:
                number, _, regex = number_line_and_regex
                message = f"line {number} matches {regex}"
                _log.info("%s: is generated code because %s", source_path, message)
                result = SourceAnalysis.from_state(source_path, group, SourceState.generated, message)
        if result is None:
            assert lexer is not None
            assert source_code is not None
            language = base_language(lexer.name) if merge_embedded_language else lexer.name
            if ("xml" in language.lower()) or (language == "Genshi"):
                dialect = pygount.xmldialect.xml_dialect(source_path, source_code)
                if dialect is not None:
                    language = dialect
            _log.info("%s: analyze as %s using encoding %s", source_path, language, encoding)
            mark_to_count_map = {"c": 0, "d": 0, "e": 0, "s": 0}
            is_markup = is_markup_file(source_path)
            for line_parts in _line_parts(lexer, source_code, is_markup=is_markup):
                mark_to_increment = "e"
                for mark_to_check in ("d", "s", "c"):
                    if mark_to_check in line_parts:
                        mark_to_increment = mark_to_check
                mark_to_count_map[mark_to_increment] += 1
            reduced_path = source_path.split(tmp_dir)[-1].lstrip(os.sep) if tmp_dir else source_path
            result = SourceAnalysis(
                path=reduced_path,
                language=language,
                group=group,
                code=mark_to_count_map["c"],
                documentation=mark_to_count_map["d"],
                empty=mark_to_count_map["e"],
                string=mark_to_count_map["s"],
                state=SourceState.analyzed,
                state_info=None,
            )

        assert result is not None
        return result

    @property
    def path(self) -> str:
        return self._path

    @property
    def language(self) -> str:
        """
        The programming language the analyzed source code is written in; if
        :py:attr:`state` does not equal :py:attr:`SourceState.analyzed` this
        will be a pseudo language.
        """
        return self._language

    @property
    def group(self) -> str:
        """
        Group the source code belongs to; this can be any text useful to group
        the files later on. It is perfectly valid to put all files in the same
        group.

        (Note: this property is mostly there for compatibility with the
        original SLOCCount.)
        """
        return self._group

    @property
    def code_count(self) -> int:
        """number of lines containing code"""
        return self._code

    @property
    def documentation_count(self) -> int:
        """number of lines containing documentation (resp. comments)"""
        return self._documentation

    @property
    def empty_count(self) -> int:
        """
        number of empty lines, including lines containing only white space,
        white characters or white code words

        See also: :py:func:`white_characters`, :py:func:`white_code_words`
        """
        return self._empty

    @property
    def line_count(self) -> int:
        """number of total lines, which is what you text editor a `wc -l`
        would show
        """
        return self.code_count + self.documentation_count + self.empty_count + self.string_count

    @property
    def string_count(self) -> int:
        """number of lines containing only strings but no other code"""
        return self._string

    @property
    def source_count(self) -> int:
        """number of source lines of code (the sum of code_count and string_count)"""
        return self.code_count + self.string_count

    @property
    def state(self) -> SourceState:
        """
        The state of the analysis after parsing the source file.
        """
        return self._state

    @property
    def state_info(self) -> Optional[Union[str, Exception]]:
        """
        Possible additional information about :py:attr:`state`:

        * :py:attr:`SourceState.duplicate`: path to the original source file
          the :py:attr:`path` is a duplicate of
        * :py:attr:`SourceState.error`: the :py:exc:`Exception` causing the
          error
        * :py:attr:`SourceState.generated`: a human-readable explanation why
          the file is considered to be generated
        """
        return self._state_info

    @property
    def is_countable(self) -> bool:
        """
        ``True`` if source counts can be counted towards a total.
        """
        return self.state in (SourceState.analyzed, SourceState.duplicate)

    def __repr__(self):
        name_to_value_map = {
            "path": repr(self.path),
            "language": repr(self.language),
            "group": repr(self.group),
            "state": self.state.name,
        }
        if self.state == SourceState.analyzed:
            name_to_value_map.update(
                {
                    "code_count": self.code_count,
                    "documentation_count": self.documentation_count,
                    "empty_count": self.empty_count,
                    "string_count": self.string_count,
                }
            )
        if self.state_info is not None:
            name_to_value_map["state_info"] = repr(self.state_info)
        return mapped_repr(self, name_to_value_map)


class SourceScanner:
    """
    Scanner for source code files matching certain conditions.
    """

    def __init__(
        self,
        source_patterns,
        suffixes="*",
        folders_to_skip=None,
        name_to_skip=None,
    ):
        self._source_patterns = source_patterns
        self._suffixes = pygount.common.regexes_from(suffixes)
        self._folder_regexps_to_skip = (
            folders_to_skip
            if folders_to_skip is not None
            else pygount.common.regexes_from(DEFAULT_FOLDER_PATTERNS_TO_SKIP_TEXT)
        )
        self._name_regexps_to_skip = (
            name_to_skip
            if folders_to_skip is not None
            else pygount.common.regexes_from(DEFAULT_NAME_PATTERNS_TO_SKIP_TEXT)
        )
        self._git_storages = []

    def close(self):
        for git_storage in self._git_storages:
            git_storage.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    @property
    def source_patterns(self):
        return self._source_patterns

    @property
    def suffixes(self) -> list[Pattern]:
        return self._suffixes

    @property
    def folder_regexps_to_skip(self) -> list[Pattern]:
        return self._folder_regexps_to_skip

    @folder_regexps_to_skip.setter
    def folder_regexps_to_skip(self, regexps_or_pattern_text):
        self._folder_regexps_to_skip.append = pygount.common.regexes_from(
            regexps_or_pattern_text, self.folder_regexps_to_skip
        )

    @property
    def name_regexps_to_skip(self) -> list[Pattern]:
        return self._name_regexps_to_skip

    @name_regexps_to_skip.setter
    def name_regexps_to_skip(self, regexps_or_pattern_text):
        self._name_regexps_to_skip = pygount.common.regexes_from(regexps_or_pattern_text, self.name_regexps_to_skip)

    def _is_path_to_skip(self, name, is_folder) -> bool:
        assert os.sep not in name, f"name={name!r}"
        regexps_to_skip = self._folder_regexps_to_skip if is_folder else self._name_regexps_to_skip
        return any(path_name_to_skip_regex.match(name) is not None for path_name_to_skip_regex in regexps_to_skip)

    def _paths_and_group_to_analyze_in(self, folder, group, tmp_dir) -> PathData:
        assert folder is not None
        assert group is not None

        for name in os.listdir(folder):
            path = os.path.join(folder, name)
            if not os.path.islink(path):
                is_folder = os.path.isdir(path)
                if self._is_path_to_skip(os.path.basename(path), is_folder):
                    _log.debug("skip due to matching skip pattern: %s", path)
                elif is_folder:
                    yield from self._paths_and_group_to_analyze_in(path, group, tmp_dir)
                else:
                    yield PathData(source_path=path, group=group, tmp_dir=tmp_dir)

    def _paths_and_group_to_analyze(self, path_to_analyse_pattern, group=None, tmp_dir=None) -> Iterator[PathData]:
        for path_to_analyse in glob.glob(path_to_analyse_pattern):
            if os.path.islink(path_to_analyse):
                _log.debug("skip link: %s", path_to_analyse)
            else:
                is_folder = os.path.isdir(path_to_analyse)
                if self._is_path_to_skip(os.path.basename(path_to_analyse), is_folder):
                    _log.debug("skip due to matching skip pattern: %s", path_to_analyse)
                else:
                    actual_group = group
                    if is_folder:
                        if actual_group is None:
                            actual_group = os.path.basename(path_to_analyse)
                            if actual_group == "":
                                # Compensate for trailing path separator.
                                actual_group = os.path.basename(os.path.dirname(path_to_analyse))
                        yield from self._paths_and_group_to_analyze_in(path_to_analyse_pattern, actual_group, tmp_dir)
                    else:
                        if actual_group is None:
                            actual_group = os.path.dirname(path_to_analyse)
                            if actual_group == "":
                                actual_group = os.path.basename(os.path.dirname(os.path.abspath(path_to_analyse)))
                        yield PathData(source_path=path_to_analyse, group=actual_group, tmp_dir=tmp_dir)

    def _source_paths_and_groups_to_analyze(self, source_patterns_to_analyze) -> list[PathData]:
        assert source_patterns_to_analyze is not None

        result = []

        def _process_source_pattern(source_pattern: str):
            remote_url, revision = git_remote_url_and_revision_if_any(source_pattern)
            if remote_url is not None:
                git_storage = GitStorage(remote_url, revision)
                self._git_storages.append(git_storage)
                git_storage.extract()
                result.extend(
                    self._paths_and_group_to_analyze(git_storage.temp_folder, tmp_dir=git_storage.temp_folder)
                )
            else:
                has_url_prefix = re.match(HTTP_URL_REGEX, source_pattern)
                if has_url_prefix:
                    is_git_url = re.match(GIT_REPO_REGEX, source_pattern_to_analyze) is not None
                    if not is_git_url:
                        raise pygount.Error(
                            f'URL to git repository {source_pattern} must end with ".git" or must match the pattern '
                            f"http(s)://({'|'.join(_ALLOWED_GIT_PLATFORMS)})/<...>/<...>.git. "
                            f"For example: git@github.com:roskakori/pygount.git or "
                            f"https://github.com/roskakori/pygount.git."
                        )
                    source_pattern = source_pattern.rstrip("/")
                    _process_source_pattern(source_pattern + ".git")
                else:
                    result.extend(self._paths_and_group_to_analyze(source_pattern_to_analyze))

        # NOTE: We could avoid initializing `source_pattern_to_analyze` here by moving the `try` inside
        #  the loop, but this would incor a performance overhead (ruff's PERF203).
        source_pattern_to_analyze = None
        try:
            for source_pattern_to_analyze in source_patterns_to_analyze:
                _process_source_pattern(source_pattern_to_analyze)
        except OSError as error:
            assert source_pattern_to_analyze is not None
            raise OSError(f'cannot scan "{source_pattern_to_analyze}" for source files: {error}') from error
        result = sorted(set(result), key=lambda data: (data.source_path, data.group))
        return result

    def source_paths(self) -> Iterator[PathData]:
        """
        Paths to source code files matching all the conditions for this scanner.
        """
        source_paths_and_groups_to_analyze = self._source_paths_and_groups_to_analyze(self.source_patterns)

        for path_data in source_paths_and_groups_to_analyze:
            suffix = os.path.splitext(path_data.source_path)[1].lstrip(".")
            is_suffix_to_analyze = any(suffix_regexp.match(suffix) for suffix_regexp in self.suffixes)
            if is_suffix_to_analyze:
                yield path_data
            else:
                _log.info("skip due to suffix: %s", path_data.source_path)


_LANGUAGE_TO_WHITE_WORDS_MAP = {"batchfile": {"@"}, "python": {"pass"}, "sql": {"begin", "end"}}
for _language in _LANGUAGE_TO_WHITE_WORDS_MAP:
    assert _language.islower()


def matching_number_line_and_regex(
    source_lines: Iterator[str], generated_regexes: Sequence[Pattern], max_line_count: int = 15
) -> Optional[tuple[int, str, Pattern]]:
    """
    The first line and its number (starting with 0) in the source code that
    indicated that the source code is generated.
    :param source_lines: lines of text to scan
    :param generated_regexes: regular expressions a line must match to indicate
        the source code is generated.
    :param max_line_count: maximum number of lines to scan
    :return: a tuple of the form ``(number, line, regex)`` or ``None`` if the
        source lines do not match any ``generated_regexes``.
    """
    initial_numbers_and_lines = enumerate(itertools.islice(source_lines, max_line_count))
    matching_number_line_and_regexps = (
        (number, line, matching_regex)
        for number, line in initial_numbers_and_lines
        for matching_regex in generated_regexes
        if matching_regex.match(line)
    )
    possible_first_matching_number_line_and_regexp = list(itertools.islice(matching_number_line_and_regexps, 1))
    result = (
        possible_first_matching_number_line_and_regexp[0] if possible_first_matching_number_line_and_regexp else None
    )
    return result


def white_characters(language_id: str) -> str:
    """
    Characters that count as white space if they are the only characters in a
    line.
    """
    assert language_id is not None
    assert language_id.islower()
    return "(),:;[]{}"


def white_code_words(language_id: str) -> set[str]:
    """
    Words that do not count as code if it is the only word in a line.
    """
    assert language_id is not None
    assert language_id.islower()
    return _LANGUAGE_TO_WHITE_WORDS_MAP.get(language_id, set())


def _delined_tokens(tokens: Iterator[tuple[TokenType, str]]) -> Iterator[TokenType]:
    for token_type, token_text in tokens:
        remaining_token_text = token_text
        newline_index = remaining_token_text.find("\n")
        while newline_index != -1:
            yield token_type, remaining_token_text[: newline_index + 1]
            remaining_token_text = remaining_token_text[newline_index + 1 :]
            newline_index = remaining_token_text.find("\n")
        if remaining_token_text != "":
            yield token_type, remaining_token_text


def _pythonized_comments(tokens: Iterator[tuple[TokenType, str]]) -> Iterator[TokenType]:
    """
    Similar to tokens but converts strings after a colon (`:`) to comments.
    """
    is_after_colon = True
    for token_type, result_token_text in tokens:
        if is_after_colon and (token_type in pygments.token.String):
            result_token_type = pygments.token.Comment
        else:
            result_token_type = token_type
            if result_token_text == ":":
                is_after_colon = True
            elif token_type not in pygments.token.Comment:
                is_whitespace = len(result_token_text.rstrip(" \f\n\r\t")) == 0
                if not is_whitespace:
                    is_after_colon = False
        yield result_token_type, result_token_text


def _line_parts(lexer: pygments.lexer.Lexer, text: str, is_markup: bool = False) -> Iterator[set[str]]:
    line_marks = set()
    tokens = _delined_tokens(lexer.get_tokens(text))
    if lexer.name == "Python":
        tokens = _pythonized_comments(tokens)
    language_id = lexer.name.lower()
    white_text = " \f\n\r\t" + white_characters(language_id)
    white_words = white_code_words(language_id)
    for token_type, token_text in tokens:
        # NOTE: Pygments treats preprocessor statements as special comments.
        is_actual_comment = token_type in pygments.token.Comment and token_type not in (
            pygments.token.Comment.Preproc,
            pygments.token.Comment.PreprocFile,
        )
        if is_actual_comment:
            line_marks.add("d")  # 'documentation'
        elif token_type in pygments.token.String:
            line_marks.add("s")  # 'string'
        else:
            is_white_text = (token_text.strip() in white_words) or (token_text.rstrip(white_text) == "")
            if not is_white_text:
                line_mark = "d" if is_markup else "c"
                line_marks.add(line_mark)
        if token_text.endswith("\n"):
            yield line_marks
            line_marks = set()
    if len(line_marks) >= 1:
        yield line_marks


def check_file_handle_is_seekable(file_handle: Optional[Union[BufferedIOBase, RawIOBase]], source_path: str):
    if not file_handle.seekable():
        raise pygount.Error(f"cannot determine encoding: file handle must be seekable: {source_path}")


def encoding_for(
    source_path: str,
    encoding: str = "automatic",
    fallback_encoding: Optional[str] = None,
    file_handle: Optional[Union[BufferedIOBase, RawIOBase]] = None,
) -> str:
    """
    The encoding used by the text file stored in ``source_path``.

    The algorithm used is:

    * If ``encoding`` is ``'automatic``, attempt the following:

      1. Check BOM for UTF-8, UTF-16 and UTF-32.
      2. Look for XML prolog or magic heading like ``# -*- coding: cp1252 -*-``
      3. Read the file using UTF-8.
      4. If all this fails, use the ``fallback_encoding`` and ignore any
         further encoding errors.

    * If ``encoding`` is ``'chardet`` use :mod:`chardet` to obtain the encoding.
    * For any other ``encoding`` simply use the specified value.
    """
    assert encoding is not None

    if encoding == "automatic":
        if file_handle is None:
            with open(source_path, "rb") as source_file:
                heading = source_file.read(128)
        else:
            check_file_handle_is_seekable(file_handle, source_path)
            heading = file_handle.read(128)
            file_handle.seek(-len(heading), SEEK_CUR)
        result = None
        if len(heading) == 0:
            # File is empty, assume a dummy encoding.
            result = "utf-8"
        if result is None:
            result = next(
                (
                    encoding_for_bom
                    for bom, encoding_for_bom in _BOM_TO_ENCODING_MAP.items()
                    if heading[: len(bom)] == bom
                ),
                None,
            )
        if result is None:
            result = encoding_from_header(heading)
    elif encoding == "chardet":
        assert _detector is not None, (
            'without chardet installed, encoding="chardet" must be rejected before calling encoding_for()'
        )
        _detector.reset()
        if file_handle is None:
            with open(source_path, "rb") as source_file:
                lines = source_file.readlines()
        else:
            check_file_handle_is_seekable(file_handle, source_path)
            file_position = file_handle.tell()
            lines = file_handle.readlines()
            file_handle.seek(file_position)
        for line in lines:
            _detector.feed(line)
            if _detector.done:
                break
        result = _detector.result["encoding"]
        if result is None:
            _log.warning(
                "%s: chardet cannot determine encoding, assuming fallback encoding %s", source_path, fallback_encoding
            )
            result = fallback_encoding
    else:
        # Simply use the specified encoding.
        result = encoding
    if result is None:
        # Encoding 'automatic' or 'chardet' failed to detect anything.
        if fallback_encoding is not None:
            # If defined, use the fallback encoding.
            result = fallback_encoding
        else:
            try:
                # Attempt to read the file as UTF-8.
                if file_handle is None:
                    with open(source_path, encoding="utf-8") as source_file:
                        source_file.read()
                else:
                    check_file_handle_is_seekable(file_handle, source_path)
                    file_position = file_handle.tell()
                    file_handle.read()
                    file_handle.seek(file_position)
                result = "utf-8"
            except UnicodeDecodeError:
                # UTF-8 did not work out, use the default as last resort.
                result = DEFAULT_FALLBACK_ENCODING
            _log.debug("%s: no fallback encoding specified, using %s", source_path, result)

    assert result is not None
    return result


def encoding_from_header(header: bytes) -> Optional[str]:
    ascii_header = header.decode("ascii", errors="replace")
    result = encoding_from_possible_magic_comment(ascii_header)
    if result is None:
        result = encoding_from_possible_xml_prolog(ascii_header)
    return result


def encoding_from_possible_magic_comment(ascii_header: str) -> Optional[str]:
    return next(_magic_comment_encodings(ascii_header), None)


def _magic_comment_encodings(ascii_header: str) -> Iterator[str]:
    header_lines = ascii_header.split("\n")[:2]
    for header_line in header_lines:
        for magic_line_start_regex in _MAGIC_COMMENT_LINE_START_REGEXES:
            magic_line_start_match = re.match(magic_line_start_regex, header_line)
            if magic_line_start_match is not None:
                remainder = magic_line_start_match.group("remainder")
                for magic_coding_comment_regex in _MAGIC_COMMENT_LINE_REMAINDER_REGEXES:
                    result = magic_coding_comment_regex.match(remainder)
                    if result is not None:
                        yield result.group("encoding")


def encoding_from_possible_xml_prolog(ascii_header: str) -> Optional[str]:
    header_line = ascii_header.replace("\f\n\r\v", " ")
    xml_prolog_match = _XML_PROLOG_REGEX.match(header_line)
    return xml_prolog_match.group("encoding") if xml_prolog_match is not None else None


def is_binary_file(source_path: str) -> bool:
    with open(source_path, "rb") as source_file:
        initial_bytes = source_file.read(8192)
    return not any(initial_bytes.startswith(bom) for bom in _TEXT_BOMS) and b"\0" in initial_bytes


def is_plain_text(source_path):
    return _PLAIN_TEXT_NAME_REGEX.match(os.path.basename(source_path))


def has_lexer(source_path: str) -> bool:
    """
    Initial quick check if there is a lexer for ``source_path``. This removes
    the need for calling :py:func:`pygments.lexers.guess_lexer_for_filename()`
    which fully reads the source file.
    """
    result = bool(pygments.lexers.find_lexer_class_for_filename(source_path))
    if not result:
        suffix = os.path.splitext(os.path.basename(source_path))[1].lstrip(".")
        result = suffix in _SUFFIX_TO_FALLBACK_LEXER_MAP
    return result


def guess_lexer(source_path: str, text: str) -> pygments.lexer.Lexer:
    if is_plain_text(source_path):
        result = pygount.lexers.PlainTextLexer()
    else:
        try:
            result = pygments.lexers.guess_lexer_for_filename(source_path, text)
        except pygments.util.ClassNotFound:
            suffix = os.path.splitext(os.path.basename(source_path))[1].lstrip(".")
            result = _SUFFIX_TO_FALLBACK_LEXER_MAP.get(suffix)
    return result


def base_language(language: str) -> str:
    base_language_match = _BASE_LANGUAGE_REGEX.match(language)
    return language if base_language_match is None else base_language_match.group("base_language")
